#include <Wire.h>
#include <LiquidCrystal_I2C.h>

#include "SoftwareSerial.h"
#include <Keypad.h>


#define Start_Byte        0x7E
#define Version_Byte      0xFF
#define Command_Length    0x06
#define End_Byte          0xEF

//  Returns info with command 0x41 [0x01: info, 0x00: no info]
#define Acknowledge       0x01


//Globale Variablen für MP3-card & Lautsprecher:
SoftwareSerial mySerial(10, 11); 
byte receive_buffer[10] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
char data;
byte volume = 0x20;
bool mute_state = false;


//Folgende Funktionen, wurden aus dem Internet entnommen
//wurden teilweise leicht überarbeitet  
//Konsolenausgaben (Serial.print) wurden auskommentiert

//Großteil der folgenden Funktionen werden durch eine weitere Funktion "befehl()" aufgerufen (siehe ganz unten)

void execute_CMD(byte CMD, byte Par1, byte Par2) {
  //  Sends the command to the module

  //  Calculate the checksum (2 bytes)
  word checksum = -(Version_Byte + Command_Length + CMD + Acknowledge + Par1 + Par2);

  //  Build the command line
  byte Command_line[10] = { Start_Byte, Version_Byte,
                            Command_Length, CMD, Acknowledge,
                            Par1, Par2, highByte(checksum),
                            lowByte(checksum), End_Byte
                          };

  //  Send the command line to the module
  for (byte i = 0; i < 10; i++) {
    mySerial.write(Command_line[i]);
  }
}

void reset_rec_buf() {
  // Resets the receive_buffer
  for (uint8_t i = 0; i < 10; i++) {
    receive_buffer[i] = 0;
  }
}

bool receive() {
  // Reads the received data
  reset_rec_buf();
  if (mySerial.available() < 10) {
    return false;
  }
  for (uint8_t i = 0; i < 10; i++) {
    short b = mySerial.read();
    if (b == -1) {
      return false;
    }
    receive_buffer[i] = b;
  }

  // when you reset the module in software,
  // receive_buffer elements are shifted
  // to correct that we do the following:
  short b = receive_buffer[0];
  for (uint8_t i = 0; i < 10; i++) {
    if (i == 9) {
      receive_buffer[i] = b;
    }
    else {
      receive_buffer[i] = receive_buffer[i + 1];
    }
  } //  end correcting receive_buffer
  return true;
}

void print_received(bool print_it) {
  // Outputs the returend data from the module
  // To the Serial Monitor
  if (print_it) {
    if (receive()) {
      for (uint8_t i = 0; i < 10; i++) {
        //Serial.print(receive_buffer[i], HEX);
        //Serial.print("\t");
      }
      //Serial.println();
    }
  }
  else {
    receive();
  }
}

void module_init() {
  // Resets the module and sets the EQ state,
  // level, and plays first song on the
  // storage device
  execute_CMD(0x0C, 0, 0); // reset the module
  delay(100);
  print_received(false);
  delay(100);
  //Serial.print("SDON\t");
  print_received(true);
  delay(100);
  set_eq(2);
}

void set_eq(uint8_t eq) {
  // Sets the state of EQ
  //Serial.print("SETEQ\t");
  execute_CMD(0x07, 0, eq); // Sets the EQ
  delay(100);
  print_received(false);
  delay(100);
  execute_CMD(0x44, 0, 0);  // Get EQ state
  delay(100);
  print_received(false);
  delay(100);
  print_received(true);
  delay(100);
}

void play_first() {
  // Plays first song on the storage device
  //Serial.print("PLYFST\t");
  execute_CMD(0x03, 0, 1); // Play first song
  delay(100);
  print_received(false);
  delay(100);
  execute_CMD(0x45, 0, 0); // Get playback status
  delay(100);
  print_received(false);
  delay(100);
  print_received(true);
  delay(100);
}

void set_volume(uint8_t volume) {
  // Sets the volume level
  //Serial.print("SETVOL\t");
  execute_CMD(0x06, 0, volume); //  Set volume level
  delay(100);
  print_received(false);
  delay(100);
  execute_CMD(0x43, 0, 0);  //  Get volume level
  delay(100);
  print_received(false);
  delay(100);
  print_received(true);
  delay(100);
}

void play() {
  // Resumes playing current song
  //Serial.print("PLAY\t");
  execute_CMD(0x0D, 0, 0);
  delay(100);
  print_received(false);
  delay(100);
  execute_CMD(0x4C, 0, 0); // Get the current song played
  delay(100);
  print_received(false);
  delay(100);
  print_received(true);
  delay(100);
}

void pause() {
  // Pauses only current song
  //Serial.print("PAUSE\t");
  execute_CMD(0x0E, 0, 0);
  delay(100);
  print_received(false);
  delay(100);
  execute_CMD(0x4C, 0, 0); // Get the current song played
  delay(100);
  print_received(false);
  delay(100);
  print_received(true);
  delay(100);
}

void play_next() {
  // Plays next song, after which it stops playback
  //Serial.print("NEXT\t");
  execute_CMD(0x01, 0, 0);
  delay(100);
  print_received(false);
  delay(100);
  execute_CMD(0x4C, 0, 0); // Get the current song played
  delay(100);
  print_received(false);
  delay(100);
  print_received(true);
  delay(100);
}

void play_previous() {
  // Plays previous song, after which it stops playback
  //Serial.print("PRE\t");
  execute_CMD(0x02, 0, 0);
  delay(100);
  print_received(false);
  delay(100);
  execute_CMD(0x4C, 0, 0); // Get the current song played
  delay(100);
  print_received(false);
  delay(100);
  print_received(true);
  delay(100);
}

void mute() {
  mute_state = !mute_state;

  if (mute_state) {
    execute_CMD(0x43, 0, 0);  //  Get volume level
    delay(100);
    print_received(false);
    delay(100);
    print_received(false);
    delay(100);
    volume = receive_buffer[6];

    //Serial.print("MUTE\t");
    execute_CMD(0x06, 0, 0x00); //  Set volume level
    delay(100);
    print_received(false);
    delay(100);
    execute_CMD(0x43, 0, 0);  //  Get volume level
    delay(100);
    print_received(false);
    delay(100);
    print_received(true);
    delay(100);
  }
  else {
    //Serial.print("VOL\t");
    execute_CMD(0x06, 0, volume); //  Set previous volume level
    delay(100);
    print_received(false);
    delay(100);
    execute_CMD(0x43, 0, 0);  //  Get volume level
    delay(100);
    print_received(false);
    delay(100);
    print_received(true);
    delay(100);
  }
}



void loop_fol_two_song_one() {
  // plays song 01 in the 02 folder
  execute_CMD(0x17, 2, 1);
  delay(100);
  //Serial.print("\t");
  print_received(false);
  delay(100);
  //Serial.print("Loop01\t");
  print_received(false);
  delay(100);
  execute_CMD(0x45, 0, 0); // Get playback status
  delay(100);
  print_received(false);
  delay(100);
  //Serial.print("PLYBCK\t");
  print_received(true);
  delay(100);
}

void loop_current() {
  // loops current playing song
  execute_CMD(0x19, 0, 0);
  delay(100);
  //Serial.print("LPCRRNT\t");
  print_received(false);
  delay(100);
  execute_CMD(0x45, 0, 0); // Get playback status
  delay(100);
  print_received(false);
  delay(100);
  //Serial.print("PLYBCK\t");
  print_received(true);
  delay(100);
}

void stop_loop_current() {
  // stops looping the current playing song
  execute_CMD(0x19, 0, 1);
  delay(100);
  //Serial.print("STPLP\t");
  print_received(false);
  delay(100);
  execute_CMD(0x45, 0, 0); // Get playback status
  delay(100);
  print_received(false);
  delay(100);
  //Serial.print("PLYBCK\t");
  print_received(true);
  delay(100);
}

void loop_all() {
  // loops all songs, playback one by one
  execute_CMD(0x11, 0, 1);
  delay(100);
  //Serial.print("LPALL\t");
  print_received(false);
  delay(100);
  execute_CMD(0x45, 0, 0); // Get playback status
  delay(100);
  print_received(false);
  delay(100);
  //Serial.print("PLYBCK\t");
  print_received(true);
  delay(100);
}

void stop_loop_all() {
  // stops loop all
  execute_CMD(0x11, 0, 0);
  delay(100);
  //Serial.print("NOLPA\t");
  print_received(false);
  delay(100);
  execute_CMD(0x45, 0, 0); // Get playback status
  delay(100);
  print_received(false);
  delay(100);
  //Serial.print("PLYBCK\t");
  print_received(true);
  delay(100);
}

void random_play() {
  // random plays all songs, loops all, repeats songs in playback
  execute_CMD(0x18, 0, 0);
  delay(100);
  //Serial.print("RNDM\t");
  print_received(false);
  delay(100);
  execute_CMD(0x4C, 0, 0);  // Get current song played
  delay(100);
  print_received(false);
  delay(100);
  print_received(true);
  delay(100);
}

void stop_all() {
  // stop playback
  execute_CMD(0x16, 0, 0);
  delay(100);
  //Serial.print("STOP\t");
  print_received(false);
  delay(100);
  execute_CMD(0x45, 0, 0); // Get playback status
  delay(100);
  print_received(false);
  delay(100);
  //Serial.print("PLYBCK\t");
  print_received(true);
  delay(100);
}


//Globale Variablen:

//MP3-card & Lautsprecher:
//siehe oben nach #include/ define

//Keypad:
const byte SPALTEN = 4;
const byte ZEILEN = 4;
char TASTEN[ZEILEN][SPALTEN] = {
  {'1', '2', '3', 'A'},
  {'4', '5', '6', 'B'},
  {'7', '8', '9', 'C'},
  {'*', '0', '#', 'D'}
};


byte SpaltenPIN[SPALTEN] = { 5, 4, 3, 2}; //Zuweisung digitaler Pins
byte ZeilenPIN[ZEILEN] = { 9, 8, 7, 6};
char gedrueckteTaste;

Keypad myKeypad = Keypad(makeKeymap(TASTEN), ZeilenPIN, SpaltenPIN, ZEILEN,
                         SPALTEN);    
//Display:
LiquidCrystal_I2C lcd = LiquidCrystal_I2C(0x27, 16, 2); 
                 

//2 Knöpfe:
const int BUTTON_PIN = A2; //Knopf 1 am Pin A2
const int BUTTON_PIN2 = A3;// Knopf 2 am Pin 12

//Variablen für den Knopf Zustand (ob Knopf gedrückt wurde oder nicht)
int lastButtonState;    // vorherige Knopf-Zustand
int currentButtonState; // aktueller Knopf-Zutand

int lastButtonState2;    
int currentButtonState2; 

//LED's:
const int LED_PIN = 12;


//Bewegungssensor:
const int motionPin = A1; //nimmt Wärmequellen zur Bewegungserkennung war
int pirState = LOW;   //Speichert den alten Wert von "motionPin"

//Generelle Globale Variablen:
char zahl[4];             //zum Speichern der eingegebenen Zeichen
String code;              //Zeichen von zahl[4] werden in dieser String gespeichert     
String passwort = "1234"; //Selbst erstelltes Passwort
String passwort2 = "";    //zum Einstellen eines induviduellen Passwortes
String passwort3 = "#187";//Easter-Egg Passwort -> Spielt Rick Astley ab          

//Strings die mehrfach benutzt werden für die Display-Funktionen
//String dispA = "ALARM! ALARM!";       //wurde lokalisiert 
String dispB = "";
//String dispC = "Eingegebener Code: "; //wurde lokalisiert

//Zähler Variablen
int j = 0;    //Für char zahl[4] ->ändert Feld des Array zum Speichern der eingetippten Zeichen (Keypad)
              //wird auch zum überprüfen benutzt, ob man 4 Zeichen eingetippt hat
int k = 0;    //Zähler um die LED's an/ auszuschalten (
              //alternative zu delay, ohne den Loop zu unterbrechen)
int l = 0;    //Merkt sich, ob der Aarm bereit aktiviert wurde 
              // 0-> Alarm aus, 1-> Alarm an



void setup() {

  //Serieller Monitor:
  Serial.begin(9600);
  mySerial.begin(9600);

  //Knöpfe:
  pinMode(BUTTON_PIN, INPUT_PULLUP); 
  currentButtonState = digitalRead(BUTTON_PIN);
  
  pinMode(BUTTON_PIN2, INPUT_PULLUP); 
  currentButtonState2 = digitalRead(BUTTON_PIN2);
  delay(100);

  //LED:
  pinMode(LED_PIN, OUTPUT);

  //Display:
  lcd.init();
  lcd.backlight();
  delay(250);
  lcd.noBacklight();
  delay(100);
  lcd.backlight();
  delay(100);

  //MP3-Lieder:
  //Serial.println("\nInitialization");
  module_init();
}

void loop() {
  if(l==0){
    scannen();
  }

  String dispC = "Eingegebener Code: ";
  
  print_received(true);

  boolean button1 = button();
  //Falls Knopf gedrückt wurde:
  if(button1 == true) {
      goto richtigPasswort; //somit werden die if-Schleifen, übersprungen
    } 
  
  //Falls 4 Zeichen fürs Passwort eingegeben wurden,
  //werden diese 4 Zeichen mit den Paswörtern verglichen
  if (j == 4) {
    
    //Speicherung der 4 Zeichen in eine String (um einfacher mit den Zeichen zu arbeiten)
    code = String(zahl[0]) + String(zahl[1]) + String(zahl[2]) + String(zahl[3]);
    Serial.print(dispC);
    Serial.println(code);

    //überprüft, ob Passwort stimmt
    if (code == passwort || code == passwort2) {
 
      //Falls der Not-Knopf (Beginn des Loops) gedrückt wurde,
      //wird hierher gesprungen
      richtigPasswort: 
      
      //Ausgabe, dass Passwort stimmt
      String disp1 = "Passwort ist richtig";
      String disp2 = "Stoppen des Alarms!";
      display1(disp1, disp2);
      Serial.println(disp1);
  
      //Stoppen der Musik
      data= 'S';
      befehl(data);

      //Ausschalten der LED's (falls sie noch an ist)
      digitalWrite(LED_PIN, LOW);


      //Wenn man mittels des Not-Knopfes den Alarm ausschaltet
      //darf der Code nicht den Wert: " " besitzen
      //andernfalls führt es zu Problemen 
      //->(nach ändern des Passwortes, wäre man in der folgenden while-Schleife gefangen)
      code = passwort;
      if(passwort == ""){
        code = passwort2;
      }
      
      //Zum Individualisieren der Passwörter
      while(code == passwort || code == passwort2){
        
        //wird ermöglich, wenn der 1. Knopf gedrückt wird
        lastButtonState    = currentButtonState;      
        currentButtonState = digitalRead(BUTTON_PIN); 

        if(currentButtonState == LOW && lastButtonState == HIGH) {
          eigPasswort();
          Serial.println("Passwort: "  + passwort);
          Serial.println("Passwort2: " + passwort2);
          Serial.println();

          //Löschen der Werte im char Array: zahl[4]
          j = 0;
          for(int i=0; i<4; i++){
            zahl[i]= ' ';
          }
        }
      }

      l=0;

      
      
      
    }
    else {
      //Ausgabe Passwort ist Falsch!
      String disp3 = "Passwort ist falsch";
      display1(disp3, dispB);
      Serial.println(disp3);


      //Überprüft, ob Easter-Egg aktiviert wird
      //-> Fals das Paswort zum Abschalten des Alarms den gleichen Code beinhaltet, wie diesen Easter-Egg-Code "#187",
      //wird die Rick Astley Musik niemals abgespielt
      if (code == passwort3) {
        
        //Ausgabe, dass Easter-Egg aktiviert wurde
        Serial.println();
        String disp4 = "Aktiviere Rick Roll!";
        display1(dispC, disp4);
        Serial.println(disp4);

        //Ändert den Alarm Sound zum Rick Astley Lied
        data= 'N';
        befehl(data);
        
      }

       //Löschen der Werte im char Array: zahl[4]
      Serial.println();
      j = 0;
      for(int i=0; i<4; i++){
        zahl[i]= ' ';
      }
    }
  
  
  }

  keypad(dispC);

  //ermöglicht die LED's zum Blinken zu bringen
  leuchten();
   
}

// An/ Ausschalten der LED's
void leuchten(){
  k++;      //bei jeder Wiederholung wird k um 1 erhöht
            //Ab erreichen eines bestimmten Wertes werden die LED'S:

  if(k==7500){                     //ausgeschaltet
    digitalWrite(LED_PIN, LOW);     
  }
    
  if(k==15000){                      //angeschaltet und
    digitalWrite(LED_PIN, HIGH);    
    k=0;                             //zurück gesetzt
  } 
}

void keypad(String disp){
  //Eingabe der Zeichen ins Keypad
  gedrueckteTaste = myKeypad.getKey();

  if (gedrueckteTaste) {

    //Löschen aller Zeichen aud dem Display 
    //Bei der Display2-Funktion sieht man vorherige ausgegebene Werte noch
    if(j ==0){
      display1(dispB,dispB);  //dispB überschreibt beide Zeilen mit: " ";
    }

    //Speicherung und Ausgabe der eingetippten Zeichen (Keypad)
    Serial.print(gedrueckteTaste);
    Serial.println();
    zahl[j] = gedrueckteTaste;
    display2(disp, zahl[0], zahl[1], zahl[2], zahl[3]);
    j++;    //Wenn ein Zeichen getippt wurde, geht der Zähler eins hoch. 
            //wenn der auf 4 ist, wird der Code mit dem Paswwort verglichen.
  }
}


//eigPasswort: ermöglicht das individuelle Einstellen des Passwortes
void  eigPasswort(){
  for(int i=0; i<4; i++){
    zahl[i]= ' ';
  }

  j=0; //Zähler Variable muss auf 0 gesetzt werden (siehe folgende while-Schleife)
  String dispD = "Neuer Code:";
  Serial.println(dispD);
  display1(dispB, dispB);
  display2(dispD, zahl[0], zahl[1], zahl[2], zahl[3]);  
  while(j<4){
    keypad(dispD);  //wenn Keypad nicht im bestimmten Zeitabschnitt betätigt wird, wird "quasi" dieser Funktionsaufruf übersprungen
                    //nur wenn ein Zeichen auf den Keypad eingetippt wird, wird 'j' um ein erhöht.
  }
  
  //Eingetippter Code, wird als Passwort gespeichert
  String altPassw = passwort2;
  passwort2 = String(zahl[0]) +String(zahl[1]) +String(zahl[2]) +String(zahl[3]);

  if(passwort2 == passwort || passwort2 == altPassw){
    String dispE = "Gleicher Code! ";
    display1(dispE, dispB);
    Serial.println(dispE);
  }
  passwort  = ""; //altes  Passwort wird gelöscht
}

boolean button(){
  //Überprüft 2. Knopf Zustand
  lastButtonState2    = currentButtonState2;      
  currentButtonState2 = digitalRead(BUTTON_PIN2); 

  //Falls Knopf gedrückt wurde:
  if(currentButtonState2 == LOW && lastButtonState2 == HIGH) {
      return true;
    } 
  return false;
}


//Display 1: 2 Strings werden im Display ausgegeben
void display1(String disp1, String disp2){
  lcd.init();           //löscht die bereits existierenden Zeichen 
                        //(fehlt bei display2(), damit es nicht flackert, wenn neue Zeichen eingetippt werden) 
  lcd.setCursor(0, 0);
  lcd.print(disp1);
  lcd.setCursor(0, 1);
  lcd.print(disp2);
}

//Display 2: 1 String und der Code mit 4 char (Passwort) werden im Display ausgegeben
void display2(String disp1, char a, char b, char c, char d){
  lcd.setCursor(0, 0);
  lcd.print(disp1);
  lcd.setCursor(0, 1);
  lcd.print(String(a) +String(b) +String(c) +String(d)); 
}

void scannen(){
  String disp5 = "Scanne";
  display1(disp5, dispB);
  Serial.println(disp5);
  
  while(l==0){
    if (digitalRead(motionPin) == HIGH) {
      if (pirState == LOW) {
        Serial.println("Bewegung erkannt");
        pirState = HIGH;
        l=1; //damit beim Loop diese Scannen Methode übersprungen wird
             //wenn der Roboter "neugestartet" wird, wird 'l' wieder auf 0 zurückgesetzt

        alarm();  //wenn Bewegung erkannt wird, wird der Alarm aktiviert
      }
    }
    else {
      if (pirState == HIGH){
        Serial.println("Keine Bewegung");
        pirState = LOW;
      }
    }
  }
}

//Aktiviert Alarm
void alarm(){
  //ALARM! ALARM! wird auf dem Display ausgegeben
  String dispA = "ALARM! ALARM!";
  display1(dispA, dispB);
  Serial.println(dispA);
  Serial.println();
  
  //Alarm-MP3 wird 
  play_first();      //abgepielt 
  set_volume(0x10);
  data= 'C';         
  befehl(data);      //geloopt 
}

//Alle Funktionenaufrufe für die MP3 Lieder
//Waren ursprünglich mehr. 
//Wir haben überflüssige Funktionen drin gelassen, falls wir die noch mal benötigen

//Befehlfunktion, die einen weiteren Funktionaufruf durchführt
void befehl(char data) {
  if (data == 'P') {
    Serial.println("\nPlay the song");
    play();
  }
  else if (data == 'p') {
    Serial.println("\nPause the song");
    pause();
  }
  else if (data == 'N') {
    Serial.println("\nPlay next song");
    play_next();
  }
  else if (data == 'R') {
    Serial.println("\nPlay previous song\t");
    play_previous();
  }
  else if (data == 'M') {
    Serial.println("\nMute/Unmute");
    mute();
  }
  else if (data == 'C') {
    //Serial.println("\nLoop one");
    loop_current();
  }
  else if (data == 'c') {
    Serial.println("\nStop loop one");
    stop_loop_current();
  }
  else if (data == 'L') {
    Serial.println("\nLoop all");
    loop_all();
  }
  else if (data == 'l') {
    Serial.println("\nStop loop all");
    stop_loop_all();
  }
  else if (data == 'S') {
    Serial.println("\nStop playback");
    stop_all();
  }
}
